/* C = mincut(X,dir)
 *
 * Computes the minimum cut from one edge of the matrix to the other
 *
 * Inputs:
 *   X:  Evalutations of 2d function to be cut along local minima
 *   dir: 0 = vertical cut, 1 = horizontal cut
 *
 *Outputs:
 *   C: Matrix containing entries indicating side
 *       -1: left (or top) side of cut
 *        0: along the cut
 *        1: right (or bottom) side of cut
 *
 * Author: Jinwei Gu  Last Modified at: 2006-12-28
 *
 * Compared with mincut.m, this seems does not accelerate much.
 * The bottleneck is not here.
 */
 
#include "mex.h" 
#include "math.h"
#include "stdio.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) 
{ 

	double* X0;
    double* X;
    double* E;
	double* C;
    double* C0;
    double cost, t1, t2, t3;
	int M, N, height, width, idx; 
	int i, j, m, n; 
	int dir;
    
	X0 = mxGetPr(prhs[0]); 
	M = mxGetM(prhs[0]); 
	N = mxGetN(prhs[0]);
    if(nrhs>1)
    	dir = mxGetScalar(prhs[1]);
	else
        dir = 0; // default
    
    X = (double*)malloc(sizeof(double)*M*N);
    if(dir==0)
    {
        height = M;
        width = N;
        for(i=0; i<height; i++)
        for(j=0; j<width; j++)
            X[i*width+j] = X0[j*M+i];
    }
    else
    {
        height = N;
        width = M;
        for(i=0; i<height; i++)
        for(j=0; j<width; j++)
            X[i*width+j] = X0[i*M+j];
    }
   
    // Allocate the current cost array, and set first row to first row of X
    E = (double*)malloc(sizeof(double)*height*width);
    for(j=0; j<width; j++)
        E[j] = X[j];

    //Starting with the second array, compute the path costs until the end
    for(i=1; i<height; i++)
    {    
        E[i*width] = X[i*width] + min(E[(i-1)*width], E[(i-1)*width+1]);

        for(j=1; j<width-1; j++)
            E[i*width+j] = X[i*width+j] + min( E[(i-1)*width+j-1], min(E[(i-1)*width+j], E[(i-1)*width+j+1]) );
    
        E[i*width+width-1] = X[i*width+width-1] + min( E[(i-1)*width+width-2], E[(i-1)*width+width-1] );
    }
    free(X);
    
    
    //Backtrace to find the cut
    C = (double*)malloc(sizeof(double)*height*width);
    memset(C, 0, sizeof(double)*height*width);
    
    cost = 1.0e+10;
    for(j=0; j<width; j++)
    {
        if(E[(height-1)*width+j]<cost)
        {
            idx = j;
            cost = E[(height-1)*width+j];
        }
    }
    for(j=0; j<width; j++)
    {
        if(j<idx)
            C[(height-1)*width+j]=-1;
        else if(j>idx)
            C[(height-1)*width+j]=+1;
        else
            C[(height-1)*width+j]=0;
    }
       
    for(i=height-2; i>=0; i--)
    for(j=0; j<width; j++)
    {
        if(idx==0)
        {
            t1 = E[i*width+idx];
            t2 = E[i*width+idx+1];
            if(t2<t1)
                idx++;
        }
        else if(idx==width-1)
        {
            t1 = E[i*width+idx];
            t2 = E[i*width+idx-1];
            if(t2<t1)
                idx--;
        }
        else
        {
            t1 = E[i*width+idx];
            t2 = E[i*width+idx-1];
            t3 = E[i*width+idx+1];
            cost = min(t1,min(t2,t3));
            if(t2==cost)
                idx--;
            else if(t3==cost)
                idx++;
        }
        
        
        for(m=0; m<width; m++)
        {
            if(m<idx)
                C[i*width+m]=-1;
            else if(m>idx)
                C[i*width+m]=+1;
            else
                C[i*width+m]=0;
        }
    }
    free(E);
    
    // output
    plhs[0] = mxCreateDoubleMatrix(M, N, mxREAL); 
    C0 = mxGetPr(plhs[0]);
    if(dir == 0)
    {        
        for(i = 0; i < height; i++) 
        for(j = 0; j < width; j++)
        {
            C0[j*M+i] = C[i*width+j];
        }
    }
    else
    {
        for(i = 0; i < height; i++) 
        for(j = 0; j < width; j++)
        {
            C0[i*M+j] = C[i*width+j];
        }
    }
    
    free(C);
} 
